#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

import numpy as np
import scipy.ndimage as ndi
import scipy.optimize as opt

import doas

from . import io
from . import misc


def get_constant_emissions(time):
    if time.year == 2015 and time.month in [1,7]:
        return 754.86
    else:
        return 534.30



def simple_leastsq(time, background_mode='mean', cloud_thr=1.0, sat_noise=0.0,
        use_prior=True, plume_tracer='XCO2_BV', plume_thr=0.25,
        use_const_emissions=False):
    """\
    filename is either a file with 2d COSMO field or Level-2 data

    y[i,j] = x * H[i,j] + b

    - y: XCO2 measurements in plume
    - H: XCO2 from Berlin tracer (e.g. XCO2_BC) normalized by emissions
    - b: background (e.g. mean of "background" tracers in COSMO)

    The plume is defined where 'plume_tracer' > 'plume_thr'.

    If `use_const_emissions` use XCO2_BV in observation vector y.
    """
    info = {}

    if time >= datetime(2015,7,1):
        starttime = datetime(2015,7,1)
    else:
        starttime = datetime(2015,1,1)

    try:
        plume, clct  = io.read_field(time, starttime,
            [plume_tracer, 'CLCT'],
            input_files='2D'
        )
    except OSError as err:
        info['number of pixels'] = -1
        return np.nan, np.nan, info

    plume, clct = map(misc.spatial_binning, [plume, clct])

    # define plume / alternative: NO2 > 1e15 / plume detection
    is_plume = (plume > plume_thr) & (clct <= cloud_thr)

    info['number of pixels'] = np.sum(is_plume)

    if np.sum(is_plume) == 0:
        return np.nan, np.nan, info


    bv, bc, jv, anth, ra, gpp, bg  = io.read_field(time, starttime,
        [
         'XCO2_BV',
         'XCO2_BC', 'XCO2_JV',  'XCO2_A',
         'XCO2_RA', 'XCO2_GPP', 'XCO2_BG'
        ],
        input_files='2D'
    )

    # spatial binning 
    bv, bc, jv, anth, ra, gpp, bg = map(misc.spatial_binning, [bv, bc, jv, anth, ra, gpp, bg])

    # background signal
    b = (jv + anth + ra - gpp + bg)[is_plume]

    if background_mode == 'model':
        bg_std = 0.0
    elif background_mode == 'mean':
        bg_std = b.std()
        b = b.mean()
    else:
        raise ValueError('unknown mode for background estimation')

    # footprint
    H = bc[is_plume] / get_constant_emissions(time)

    # observation vector
    if use_const_emissions:
        y = (bc + jv + anth + ra - gpp + bg)[is_plume]
    else:
        y = (bv + jv + anth + ra - gpp + bg)[is_plume]

    # add noise
    y += sat_noise * np.random.randn(y.size)


    # Se, xa and Sa
    if sat_noise < 1e-16:
        Se = np.nan
    else:
        Se = np.full(y.shape, sat_noise**2 + bg_std**2)

    if use_prior:
        xa = np.array([534.0])
        Sa = np.array([262.0])**2
    else:
        xa = np.nan
        Sa = np.nan

    # forward model
    f = lambda x,_: H * x + b

    x, res = doas.solver.gauss_newton(np.array([534.0]), y, f, Se=Se, xa=xa, Sa=Sa)

    return x[0], np.sqrt(float(res['S'])), info






