#!/usr/bin/env python
# coding: utf-8

from __future__ import print_function, division
from six import string_types

from datetime import datetime, timedelta
import glob
import os

import numpy as np
import netCDF4
import xarray

from . import config as cfg
from . import misc


def str2char(s):
    return np.array(list(s), 'c')


def init_netcdf(filename, rlon, rlat, lon=None, lat=None, pollon=cfg.pollon, pollat=cfg.pollat,
        attrs=None):
    """\
    Init netCDF4 file with cosmo rotated pole grid
    """

    with netCDF4.Dataset(filename, 'w') as nc:

        if attrs is not None:
            nc.setncatts(attrs)

        # dimensions
        nc.createDimension('time')
        nc.createDimension('rlat', rlat.size)
        nc.createDimension('rlon', rlon.size)

        # time variable
        nc_time = nc.createVariable('time', 'f8', ('time',))
        nc_time.setncattr('units', str2char('seconds since 2015-01-01 00:00'))
        nc_time.setncattr('calendar', str2char('proleptic_gregorian'))

        # rlon
        nc_rlon = nc.createVariable('rlon', 'f8', ('rlon',))
        nc_rlon[:] = rlon
        nc_rlon.setncattr('axis', str2char('X'))
        nc_rlon.setncattr('units', str2char('degrees'))
        nc_rlon.setncattr('standard_name', str2char('longitude'))

        # rlat
        nc_rlat = nc.createVariable('rlat', 'f8', ('rlat',))
        nc_rlat[:] = rlat
        nc_rlat.setncattr('axis', str2char('Y'))
        nc_rlat.setncattr('units', str2char('degrees'))
        nc_rlat.setncattr('standard_name', str2char('latitude'))

        # rotated pole
        nc_pole = nc.createVariable('rotated_pole', 'c', ('time','rlat','rlon'))
        nc_pole.setncattr('grid_mapping_name', str2char('rotated_latitude_longitude'))
        nc_pole.setncattr('grid_north_pole_latitude', pollat)
        nc_pole.setncattr('grid_north_pole_longitude', pollon)
        nc_pole.setncattr('north_pole_grid_longitude', 0.0)

        # longitude
        if lon is not None:
            nc_lon = nc.createVariable('lon', 'f8', ('rlat', 'rlon'))
            nc_lon[:] = lon
            nc_lon.setncattr('axis', str2char('X'))
            nc_lon.setncattr('units', str2char('degrees'))
            nc_lon.setncattr('standard_name', str2char('longitude'))

        # latitude
        if lat is not None:
            nc_lat = nc.createVariable('lat', 'f8', ('rlat', 'rlon'))
            nc_lat[:] = lat
            nc_lat.setncattr('axis', str2char('Y'))
            nc_lat.setncattr('units', str2char('degrees'))
            nc_lat.setncattr('standard_name', str2char('latitude'))



def append_variable(filename, name, values, attrs=None):
    """\
    Append variable (name, values) to existing file.
    """
    if attrs is None:
        attrs = {}

    with netCDF4.Dataset(filename, 'a') as nc:
        var = nc.createVariable(name, values.dtype, ('time', 'rlat', 'rlon'), zlib=True)
        var[0] = values

        # add or update attributes
        for key, value in attrs.items():
            if isinstance(value, string_types):
                var.setncattr(key, str2char(value))
            else:
                var.setncattr(key, value)


def berlin_slice():
    return slice(0,1), slice(290,320), slice(330,370)


def jaenschwalde_slice():
    return slice(0,1), slice(242-15,242+15), slice(416-15,416+15)



def read_field(time, starttime, keys, indices=Ellipsis, input_files='3D',
        path='/project/d73/output/production_run_02'
    ):
    if indices is Ellipsis:
        indices = [Ellipsis for _ in keys]


    if input_files in ['3Dc', '3D']:
        hstart = (time - starttime).days // 3 * 72
        hstop = hstart + 72

        folder = '%s_%d_%d' % (starttime.strftime('%Y%m%d%H'), hstart, hstop)

        if input_files == '3D':
            filename = os.path.join(
                    path, folder, 'cosmo_output',
                    time.strftime('lffd%Y%m%d%H.nc')
            )
        else:
            filename = os.path.join(
                    path, folder, 'cosmo_output',
                    starttime.strftime('lffd%Y%m%d%Hc.nc')
            )

    elif input_files == '2D':
        filename = os.path.join(
                path, time.strftime('cosmo_2d_%Y%m%d%H.nc')
        )



    with netCDF4.Dataset(filename) as nc:
        data = [nc.variables[key][index] for key, index in zip(keys, indices)]

    return data[0] if len(data) == 1 else data


class SmartcarbOutput(object):
    def __init__(self, path):
        self.path = path

        # get starttime and file suffixes
        self.suffixes = []
        starttimes = []
        for filename in glob.glob(os.path.join(path, 'lffd??????????c*.nc')):
            filename = os.path.basename(filename)
            starttimes.append(datetime.strptime(filename[4:14], '%Y%m%d%H'))
            self.suffixes.append(filename[15:-3])

        if len(set(starttimes)) > 1:
            raise RuntimeError('')
        self.starttime = starttimes[0]

        # hstart and hstop time
        self.hstart_time, self.hstop_time = self.get_interval()

        # create a mapping for keys to filename pattern
        # Warning: Some variable names appear in more than one file.
        #          (assuming duplicates have same values!)
        self.name2pattern = {}
        for suffix in self.suffixes:

            # constant fields
            pattern = self.starttime.strftime(
                    'lffd%%Y%%m%%d%%Hc%s.nc' % suffix
            )
            filename = os.path.join(self.path,  pattern)
            with netCDF4.Dataset(filename, 'r') as nc:
                for key in nc.variables.keys():
                    self.name2pattern[key] = pattern

            # time series
            pattern = 'lffd%%Y%%m%%d%%H%s.nc' % suffix
            filename = os.path.join(
                    self.path, self.hstart_time.strftime(pattern)
            )
            with netCDF4.Dataset(filename, 'r') as nc:
                for key in nc.variables.keys():
                    self.name2pattern[key] = pattern


    def get_interval(self):
        suffix = self.suffixes[0]
        pattern = 'lffd%%Y%%m%%d%%H%s.nc' % suffix

        filenames = sorted(glob.glob(
            os.path.join(self.path, 'lffd??????????%s.nc' % suffix))
        )
        hstart = datetime.strptime(os.path.basename(filenames[0]), pattern)
        hstop = datetime.strptime(os.path.basename(filenames[-1]), pattern)

        return hstart, hstop



    def read_variable(self, name, time=None):
        """
        """
        if time is None:
            time = self.starttime

        filename = os.path.join(
                self.path,
                time.strftime(self.name2pattern[name])
        )
        with netCDF4.Dataset(filename, 'r') as nc:
            var = nc.variables[name][...]

        return var


    def read_timeseries(self, name, start=None, stop=None):
        if start is None:
            start = self.hstart_time
        if stop is None:
            stop = self.hstop_time


        return np.concatenate([self.read_variable(name, time)
            for time in misc.iter_times(start, stop)
        ])



    def iter_times(self, tracers=None, start=None, stop=None, step=None):

        if start is None:
            start = self.hstart_time
        if stop is None:
            stop = self.hstop_time
        if step is None:
            step = timedelta(hours=1)

        if tracers is None:
            tracers = []

        # constant fields
        lat = self.read_variable('lat', self.hstart_time)
        lon = self.read_variable('lon', self.hstart_time)

        rlat = self.read_variable('rlat', self.hstart_time)
        rlon = self.read_variable('rlon', self.hstart_time)

        h = self.read_variable('HHL')

        date = start

        while date < stop:

            # time-dependent fields
            p = self.read_variable('P', date)
            T = self.read_variable('T', date)
            ps = self.read_variable('PS', date)
            clct = self.read_variable('CLCT', date)

            xm = dict((name, self.read_variable(name, date))
                for name in tracers
            )
            yield date, lon, lat, rlon, rlat, h, T, p, ps, xm, clct

            date = date + step




if __name__ == '__main__':
    pass


