#!/usr/bin/env python
# coding: utf-8

from datetime import datetime, timedelta

from collections import defaultdict
import os
import time

import cartopy.crs as ccrs
from netCDF4 import Dataset

import numpy as np
import shapely

from smartcarb import orbit

from . import config as cfg
from . import io


np.seterr(divide='ignore', invalid='ignore')



def get_tracer_names(instrument, tracer_names, variable_names):
    return tracer_names



def read_albedo(filename, utc, names):
    """\
    Read NIR and SWIR albedo from file.
    """
    t = (utc - datetime(2015,1,1)).total_seconds()

    fields = []
    attrs = []

    with Dataset(filename) as nc:
        index = np.argmin(np.abs(nc.variables['time'][:] - t))
        print(index)

        for name in names:
            v = nc.variables[name]
            fields.append(v[index])
            attrs.append({
                'standard_name': v.standard_name,
                'long_name':     v.long_name,
                'units':         v.units
            })

    return fields, attrs



def iter_pixel_area(pixel, x0=cfg.startlon, y0=cfg.startlat,
        dx=cfg.dlon, dy=cfg.dlat, ie=cfg.ie, je=cfg.je):
    """\
    Iterate over indices of grid cells and yield fraction of overlap
    of pixel and cell.
    """

    if not pixel.intersects(cfg.DOMAIN):
        raise StopIteration

    dx2, dy2 = 0.5 * dx, 0.5 * dy

    # polygon bounds
    minx, miny, maxx, maxy = pixel.bounds

    # round up and down
    minx = np.floor(minx / dx) * dx
    miny = np.floor(miny / dy) * dy
    maxx = np.ceil(maxx / dx) * dx + 0.5 * dx
    maxy = np.ceil(maxy / dy) * dy + 0.5 * dy

    # start point for pixel
    i0 = int(round((minx - x0) / dx))
    j0 = int(round((miny - y0) / dy))

    # iter over grid cells
    for i,x in enumerate(np.arange(minx,maxx,dx), i0):
        for j,y in enumerate(np.arange(miny,maxy,dy), j0):

            if 0 <= i < ie and 0 <= j < je:
                grid_cell = shapely.geometry.Polygon([
                    [x-dx2, y-dy2],
                    [x-dx2, y+dy2],
                    [x+dx2, y+dy2],
                    [x+dx2, y-dy2]
                ])

                # calculate fraction of intersection and yield
                if grid_cell.intersects(pixel):
                    weight = grid_cell.intersection(pixel).area / grid_cell.area
                    yield j, i, weight



def create_pixel(lon, lat, pollon=cfg.pollon, pollat=cfg.pollat):
    """\
    Create pixel polygon in rotpole coords from corner longitude
    and latitude.
    """
    inp = ccrs.PlateCarree()
    out = ccrs.RotatedPole(
        pole_latitude=pollat,
        pole_longitude=pollon
    )
    coords = out.transform_points(inp, lon, lat)
    pixel = shapely.geometry.Polygon(coords)

    return pixel


def sum_satellite_pixels(fields, lon, lat):
    """\
    Create satellite observations from orbit filename.

    fields   :: data fields as list of 2D/3D arrays
    lon, lat :: longitude and latitude of satellite pixel corners
                with shape: (nobs, npixel, ncorners)
    """
    n_fields = len(fields)
    shape = lon.shape[:2]

    values = np.zeros((n_fields, shape[0], shape[1]), dtype='f4')
    weights = np.zeros(shape, dtype='f4')

    for n,m in np.ndindex(shape):

        pixel = create_pixel(lon[n,m], lat[n,m])

        for j,i,weight in iter_pixel_area(pixel):
            for k in range(n_fields):

                if np.ndim(fields[k]) == 3:
                    values[k,n,m] += fields[k][0,j,i] * weight
                else:
                    values[k,n,m] += fields[k][j,i] * weight

                weights[n,m] += weight

    values /= weights[np.newaxis]
    values = np.ma.array(values, mask=np.isnan(values))

    return values



def read_orbit(orbit_filename):
    with Dataset(orbit_filename) as nc:
        time = nc.variables['time'][...]
        lon0 = nc.variables['longitude'][...]
        lat0 = nc.variables['latitude'][...]
        lonc = nc.variables['longitude_corners'][...]
        latc = nc.variables['latitude_corners'][...]
        sza =  nc.variables['solar_zenith_angle'][...]

    return time, lon0, lat0, lonc, latc, sza


def date2orbit(date, instrument=None):
    """\
    """
    cycle_duration, cycle_length = {
                       #  days, orbits in cycle
            'Sentinel_5': (29,  412),
            'Sentinel_7': (11,  164)
    }[instrument]

    days = (date - datetime(2015,1,1)).total_seconds() / 86400.0
    return round(days / cycle_duration * cycle_length)



def calculate_sza():
    pass


def any_pixels_in_domain(lon, lat, domain=cfg.DOMAIN, domain_projection=cfg.ROTPOLE):
    """\
    Check if any (lon, lat) intersects with model domain.
    """
    lon, lat = lon.flatten(), lat.flatten()
    points = domain_projection.transform_points(cfg.WGS84, lon, lat)
    points = shapely.geometry.MultiPoint(points)
    return points.intersects(cfg.DOMAIN)


def main(orbit_path, cosmo_2d_path, output_path, tracer_names,
        overpass_utc=None, online_orbit=False, instrument=None,
        orbit_number=0, lon_eq=0.0, do_shift_orbit=False,
        albedo_path=None, albedo_nir_name='Albedo_NIR',
        albedo_swir_name='Albedo_SWIR',
    ):
    """\
    Create Level-2.

    Arguments:
    - orbit_path :: str
        netCDF file with orbit data (e.g. lon/lat of pixels)
        OR file describing orbit parameters

        If file describes orbit parameter, `online_orbit` needs
        to be `True` and `instrument`, `orbit_number` and `lon_eq`
        needs to be provided.

    - cosmo_2d_path :: str
        path to COSMO 2D file; can use datetime format for
        automatic transforming

    - output_path :: str
        path data output

    - tracer_names :: list of str
        list of tracer names that will be taken from COSMO 2D file and
        averaged on instrument pixels

    - overpass_utc :: datetime (default: None)
        explicit overpass time that will overwrite, if not None,
        time from orbit file/computation

    - online_orbit :: boolean (default: False)
        compute orbit online (requires SRON orbit simulator)

    - do_shift_orbit :: boolean (default: False)
       shift orbit using orbit number and cycle duration of given
       instrument

    - albedo_path :: str (default: None)
        path to albedo filename
    """
    # names of attributes copied from COSMO/albedo file
    attribute_names = ['long_name', 'standard_name', 'units']

    # compute or read orbit
    if online_orbit:
        orbit_data = orbit.compute_orbit(instrument, orbit_number,
                long_eq=lon_eq, do_reshape=True, write_to_file=False
        )
        times = orbit_data.times
        lon0 = orbit_data.lon_center
        lat0 = orbit_data.lat_center
        lonc = orbit_data.lon_corner
        latc = orbit_data.lat_corner
        sza = orbit_data.sza

    # shift pre-computed orbit
    elif do_shift_orbit:

        print('Warning: Shifting orbits not fully implemented.')

        # get cycle and orbit number
        if instrument == 'Sentinel_5':
            cycle_duration = 29.0 # days
            cycle_number, number  = divmod(orbit_number, 412)
        else:
            cycle_duration = 11.0 # days
            cycle_number, number  = divmod(orbit_number, 164)

        # TODO: should make sure that orbit path has a number format string
        orbit_path = orbit_path.format(number=number)
        times, lon0, lat0, lonc, latc, sza = read_orbit(orbit_path)

        # TODO: update time and SZA
        times[:] = np.nan
        sza[:] = np.nan

        # shift equator passing longitude (e.g. 0 or 180) for two satellites
        lon0 = (lon0 + lon_eq + 180.0) % 360 - 180.0
        lonc = (lonc + lon_eq + 180.0) % 360 - 180.0

        if np.any((lonc < -180.0) | (lonc > 180.0)):
            raise ValueError

    # load orbit from file
    else:
        orbit_data = None
        times, lon0, lat0, lonc, latc, sza = read_orbit(orbit_path)


    # check if any satellite pixels are in COSMO domain
    if not any_pixels_in_domain(lon0, lat0, cfg.DOMAIN, cfg.ROTPOLE):
        print('All satellite ground pixels are outside COSMO domain.')
        return

    # set overpass time and cosmo file
    if overpass_utc is None:
        overpass_utc = datetime(2015,1,1) + timedelta(seconds=np.mean(times))

    cosmo_filename = overpass_utc.strftime(cosmo_2d_path)

    if not os.path.exists(cosmo_filename):
        print('No COSMO file "%s" found for overpass time.' % cosmo_filename)
        return


    # init netCDF file
    output_filename = os.path.join(
        output_path,
            '%s_%s_o%04d_l%03d_data.nc' % (instrument, overpass_utc.strftime('%Y%m%d%H'),
                orbit_number, lon_eq)
    )
    print(output_filename)
    io.init_ncfile(output_filename, instrument, times, lon0, lat0, lonc, latc, sza, orbit_data)


    # average albedos on satellite lattice and save to output file
    if albedo_path is not None:
        fields, attrs = read_albedo(albedo_path, overpass_utc,
                [albedo_nir_name, albedo_swir_name]
        )
        fields = sum_satellite_pixels(fields, lonc, latc)

        io.add_variable(output_filename, albedo_nir_name, fields[0], attrs[0])
        io.add_variable(output_filename, albedo_swir_name, fields[1], attrs[1])


    # average COSMO fields on satellite lattice and save to output file
    with Dataset(cosmo_filename) as nc:
        tracer_names = get_tracer_names(instrument, tracer_names, nc.variables.keys())

        variables = [nc.variables[name] for name in tracer_names]
        attributes = [dict((a,v.getncattr(a)) for a in attribute_names) for v in variables]

        fields = sum_satellite_pixels(variables, lonc, latc)

        for name, field, attr in zip(tracer_names, fields, attributes):
            io.add_variable(output_filename, name, field, attr)


    return
    # add uncertainty scenarios to output file




    # open 2D fields




    # TODO: add uncertainties
    #
    #

    # check if any non-masked observations
    if np.all(satobs.mask):
        print('orbit is outside COSMO domain')
        return

    print('orbit with pixels is inside COSMO domain')

    # re-pack satellite observations
    satobs = dict((name, satobs[i]) for i,name in enumerate(names))

    # output filename
    output_filename = os.path.join(
        output_path,
            '%s_%s_o%04d_l%03d_data.nc' % (instrument, overpass_utc.strftime('%Y%m%d%H'), orbit_number, lon_eq)
    )
    save_satobs(
            output_filename, instrument, time, orbit_data, satobs,
            lon0, lat0, lonc, latc, sza
    )





def main2():
    if online_orbit:
        orbit_data = orbit.compute_orbit(instrument, orbit_number,
                long_eq=lon_eq, do_reshape=True, write_to_file=False
        )
        time = orbit_data.times
        lon0 = orbit_data.lon_center
        lat0 = orbit_data.lat_center
        lonc = orbit_data.lon_corner
        latc = orbit_data.lat_corner
        sza = None

        cycle_duration = 0.0
        cycle_number = 0

    else:
        orbit_data = None

        # get cycle and orbit number
        #if instrument == 'Sentinel_5':
        #    cycle_duration = 29.0 # days
        #    cycle_number, number  = divmod(orbit_number, 412)
        #else:
        #    cycle_duration = 11.0 # days
        #    cycle_number, number  = divmod(orbit_number, 164)

        # get lon, lat, utc_time from orbit
        #orbit_filename = os.path.join(
        #        path_orbits, instrument,
        #        '%s_o%03d_l000_pixel.nc' % (instrument, number)
        #)
        time, lon0, lat0, lonc, latc, sza = read_orbit(orbit_filename)

        # TODO: update time and SZA

        # shift equator passing longitude, e.g. 0,90,180 or 270
        # for four satellites
        #lon0 = (lon0 + lon_eq + 180.0) % 360 - 180.0
        #lonc = (lonc + lon_eq + 180.0) % 360 - 180.0

        if np.any((lonc < -180.0) | (lonc > 180.0)):
            raise ValueError


    # check if any latc, lonc is in domain
    if not (lon0.min() < 19.8 and lon0.max() > 7.5):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    # use closest COSMO time or provided time
    if utc is None:

        # seconds since 2015-01-01 00:00:00"
        utc = datetime(2015,1,1) + timedelta(seconds=np.mean(time))

        # add cycles
        utc += timedelta(days=cycle_number*cycle_duration)

        # round to next hour (TODO: use linear interpolation?)
        utc = datetime(utc.year, utc.month, utc.day) + timedelta(hours=round(utc.hour + utc.minute / 60.0))

    # open 2d fields
    cosmo_filename = os.path.join(path_2d_fields, utc.strftime('cosmo_2d_%Y%m%d%H.nc'))

    # create satellite observations
    try:
        with Dataset(cosmo_filename) as nc:
            if names is None:
                if instrument == 'Sentinel_7_CO2':
                    names = sorted([
                        key for key in nc.variables.keys()
                        if key.startswith('XCO2') or key == 'CLCT'
                    ])
                else:
                    names = sorted([
                        key for key in nc.variables.keys()
                        if key.endswith('_column') or key == 'CLCT' # TODO
                    ])

            variables = [nc.variables[name] for name in names]
            satobs = create_satobs(variables, lonc, latc)

    except OSError:
        print('failed to open "%s"' % cosmo_filename)
        return

    if np.all(satobs.mask):
        print('%s orbit %d with lon %d is outside domain' % (instrument, orbit_number, lon_eq))
        return

    print('%s orbit %d with lon %d is inside domain' % (instrument, orbit_number, lon_eq))

    satobs = dict((name, satobs[i]) for i,name in enumerate(names))

    # output filename
    output_filename = os.path.join(
            path_level2_data,
            '%s_%s_o%04d_l%03d_data.nc' % (instrument, utc.strftime('%Y%m%d%H'), orbit_number, lon_eq)
    )
    save_satobs(
            output_filename, instrument, time, orbit_data, satobs,
            lon0, lat0, lonc, latc
    )

    return utc









